#!/usr/bin/perl -w

die "Change this script to use a version of Perl greater than or equal to 5."
    if $PERL_VERSION < 5;

use strict;
use English;
use Getopt::Long;

use DBI;
use ba_mysql_shared;

use ba_harness_shared;

use constant false => 0;
use constant true => 1;

# report_metrics can be many things (list), but just vrt for now...
my @options_report_metrics = ("low:vrt:mean");
my $options_should_sort_traces = true;
my $options_sort_traces = "+num_users,-user_arrivals," .
    "+task_set_change_range," .
    "+tasks_per_task_set,+service_time,+think_time";
my $options_verbose = false;
my $options_quiet = false;
my $options_db_host = $db_default_loc;
my $options_baseline_set = "baseline.dis_sched = \"NOOP\""; # interactive
my $options_new_set = "new.dis_sched != \"NOOP\" AND " .
    "new.dis_sched != \"n/a\""; # batchactive
# e.g., "trace.user_arrivals_kind = "poisson" AND
#   trace.users_arrivals_mean > 100":
my $options_trace_restrict = "";
my $options_req_load_hack = false;
my $options_id_file = undef;

my %hack_sort_closure; # turns my stomache!

# collated: key: trace_input_id, val: list ref where each list elem is
# a parsed_output corresponding to the trace_input_id.
my $collated_baseline_set;
my $collated_new_set;
my $trace_ids;
my %parsed_traces; # indexed by trace id
# scanned_sort_traces is needed by sort_traces and possibly by an
# output print func to print trace characteristics in order.
my @scanned_sort_traces;

my $show_output_buffer;
my $show_func = \&show_print_wrapper;

# num_criteria_met is a hash keyed by a report-metric and whose val is
# the number of times its criteria was met.  it must be <= the number
# of trace ids.
my %num_criteria_met;

my @analyze_ids;

my ($total_baseline, $total_new);

######################################################################

&parse_options;
&setup;

&analyze;

print "total_baseline: $total_baseline, total_new: $total_new\n";
print "tbl / tn: " . $total_baseline / $total_new . ", tn / tbl: " .
    $total_new / $total_baseline . "\n";

exit 0;

sub setup
{
    $total_baseline = $total_new = 0;

    if($options_should_sort_traces) {
	&scan_sort_traces;
    }

    &show_settings;
    &show_force("\n");
}

######################################################################

sub analyze
{
    &db_ba_connect($ba_harness_ver, $options_db_host, "dpetrou",
		   $db_dpetrou_pw);

    &show_total_stats;
    &create_trace_ids;
    &create_collated_sets;
    &show_collated_info;
    &collect_traces;
    if($options_should_sort_traces) {
	$trace_ids = &sort_traces($trace_ids, \%parsed_traces);
    }

    &show_set_func(\&show_buffer_func);
    foreach my $trace_id (@$trace_ids) {
	my $should_output;

	$should_output =
	    &analyze_outputs_using_trace_id($trace_id,
				    $collated_baseline_set->{$trace_id},
					    $collated_new_set->{$trace_id});
	if($should_output) {
	    &show_buffer_flush;
	} else {
	    &show_buffer_clear;
	}
    }
    &show_set_func(\&show_print_wrapper);

    if($options_quiet) {
	&show_force("\n");
    } else {
	&show($line_sep);
    }
   
    &show_num_criteria_met;

    &write_id_file($options_id_file, \@analyze_ids);

    &db_disconnect;
}

sub show_total_stats
{
    my ($count1, $count2);
    
    $count1 = &db_select_trace_where_project_count;
    $count2 = &db_select_output_where_project_count;
    &show_force("total db rows: $count1 (trace), $count2 (output)\n");
    die "uh, trace or output count is 0" if($count1 == 0 || $count2 == 0);
}

sub create_trace_ids
{
    if($options_trace_restrict eq "--") {
	while(<STDIN>) {
	    chop;
	    push @$trace_ids, $_;
	}
#	close STDIN or die "pain"; this causes problems in
#	write_id_file for some reason XXX
    } else {
	$trace_ids =
	    &db_get_trace_id_with_constraints($options_baseline_set,
					      $options_new_set,
					      $options_trace_restrict);
    }
    die "constraints too restrictive; no traces!" if @$trace_ids == 0;
}

sub create_collated_sets
{
    my $baseline_set;
    my $new_set;

    $baseline_set =
	&db_select_output_where_trace_input_ids("baseline",
						$options_baseline_set,
						$trace_ids);
    die "got no output for baseline" if @$baseline_set == 0;
    $new_set = &db_select_output_where_trace_input_ids("new",
						       $options_new_set,
						       $trace_ids);
    die "got no output for new" if @$new_set == 0;
    $collated_baseline_set = &collate_set($baseline_set);
    $collated_new_set = &collate_set($new_set);
}

sub collate_set
{
    my $set = shift;
    my %collated_set;
    my $list_of_outputs;

    foreach my $output (@$set) {
	$list_of_outputs = $collated_set{$output->{trace_input_id}};
	push @$list_of_outputs, $output;
	$collated_set{$output->{trace_input_id}} = $list_of_outputs;
    }

    return \%collated_set;
}

sub show_collated_info
{
    my ($count1, $count2a, $count2b);

    ($count1, $count2a) = &show_collated_info_helper($collated_baseline_set);
    &show_force("using baseline runs: $count1\n");
    ($count1, $count2b) = &show_collated_info_helper($collated_new_set);
    &show_force("using new runs: $count1\n");
    die "both sets should reference the same # of traces"
	if $count2a != $count2b;
    &show_force("referenced traces: $count2a\n");
}

sub show_collated_info_helper
{
    my $set = shift;
    my ($count1, $count2);

    $count1 = 0;
    foreach my $id (keys %$set) {
	my $list;
	$list = $set->{$id};
	$count1 += @$list;
    }

    $count2 = keys %$set;

    return ($count1, $count2);
}

sub collect_traces
{
    foreach my $id (@$trace_ids) {
	my $parsed_trace;

	$parsed_trace = &db_select_trace_id($id);
	$parsed_traces{$id} = $parsed_trace;
    }
}

sub show_num_criteria_met
{
    &show_force("number of times the criteria was met for each metric:\n");
    foreach my $full_metric (@options_report_metrics) {
	my ($str1, $num, $frac, $str2);

	$str1 = sprintf "%-16s: ", $full_metric;
	$num = $num_criteria_met{$full_metric};
	$frac = $num / keys %$collated_baseline_set;
	$str2 = sprintf "%.3f", $frac;
	&show_force($str1 . "$num ($str2)\n");
    }
}

######################################################################

sub analyze_outputs_using_trace_id
{
    my ($trace_id, $baseline_outputs, $new_outputs) = @_;
    my $parsed_trace;
    my $should_output_trace_results;
    
    $parsed_trace = $parsed_traces{$trace_id};
    
    &show($line_sep);
    &show_trace_info($parsed_trace);
    
    $should_output_trace_results = false;
    foreach my $full_metric (@options_report_metrics) {
	my $met_criteria;
	
	$met_criteria = &analyze_outputs_metric($full_metric,
						$baseline_outputs,
						$new_outputs);
	
	if($met_criteria) {
	    $should_output_trace_results = true;
	}
    }

    return $should_output_trace_results;
}

sub show_trace_info
{
    my $parsed_trace = shift;
    my $trace_id;

    $trace_id = $parsed_trace->{id};
    &show("trace id: $trace_id\n");

    foreach my $trace_arg (@trace_args_list) {
	&show("    $trace_arg: " . $parsed_trace->{$trace_arg} . "\n");
    }
}

sub analyze_outputs_metric
{
    my ($full_metric, $baseline_outputs, $new_outputs) = @_;
    my ($order, $metric, $field, $filter_dir, $filter_amount);
    my $should_output;

    ($order, $metric, $field, $filter_dir, $filter_amount) =
	&parse_metric($full_metric);

    $should_output = &show_best_outputs($baseline_outputs,
					$new_outputs, $order,
					$metric, $field,
					$filter_dir, $filter_amount,
					$full_metric);

    return $should_output;
}

sub show_metric_line
{
    my ($order, $metric, $field) = @_;
    my ($metric_line_helper, $metric_line);

    if(defined($field)) {
	$metric_line_helper = "$field of $metric";
    } else {
	$metric_line_helper = "$metric";
    }

    $metric_line .= "metric          : $metric_line_helper " .
	"($order is better)\n";
    &show($metric_line);
}

sub show_sorted_outputs
{
    my ($baseline_outputs, $new_outputs, $order, $metric, $field) = @_;
    my $combined_outputs;
    my $i;

    if($options_verbose == false) {
	return;
    }

    $combined_outputs = [@$baseline_outputs, @$new_outputs];
    $combined_outputs = &sort_results($combined_outputs, $order, $metric,
				      $field);

    $i = 1;
    foreach my $output (@$combined_outputs) {
	my ($val, $sched_str, $id);
	my $line;

	$line = sprintf "rank %4d       : ", $i;
	$val = &get_val($output->{$metric}, $field);
	$sched_str = &make_sched_str($output);
	$id = $output->{id};
	$line .= "$val (sched: $sched_str; run: $id)\n";
	&show_verbose($line);
	$i++;
    }
}

sub show_best_outputs
{
    my ($baseline_outputs, $new_outputs, $order, $metric, $field,
	$filter_dir, $filter_amount, $full_metric) = @_;
    my ($best_baseline, $best_baseline_val);
    my ($best_new, $best_new_val);
    my $improvement;
    my $met_criteria;
    my $metric_req_load_hack;

    $best_baseline = &find_best_output($baseline_outputs, $order, $metric,
				       $field);
    $best_new = &find_best_output($new_outputs, $order, $metric, $field);

    if($options_req_load_hack) {
	die if($metric ne "req_load");
	$metric_req_load_hack = "load";
	$best_baseline = &find_best_output($baseline_outputs, $order,
					   $metric_req_load_hack, $field);
    }

    $best_baseline_val = &get_val($best_baseline->{$metric}, $field);
    $best_new_val = &get_val($best_new->{$metric}, $field);
    
    if($options_req_load_hack) {
	$best_baseline_val = &get_val($best_baseline->{$metric_req_load_hack},
				      $field);
    }
    
    $total_baseline += $best_baseline_val;
    $total_new += $best_new_val;

    $improvement = &find_improvement($best_baseline_val, $best_new_val,
				     $order);
    $met_criteria = &metric_met_criteria($improvement, $filter_dir,
					 $filter_amount);

    if($met_criteria) {
	&show("\n");
	&show_metric_line($order, $metric, $field);
	&show_sorted_outputs($baseline_outputs, $new_outputs, $order,
			     $metric, $field);
	&show_best_helper($best_baseline, $best_baseline_val, "baseline");
	&show_best_helper($best_new, $best_new_val, "new");
	&show_improvement($improvement);

	&store_analysis($best_baseline, $best_baseline_val, $best_new,
			$best_new_val, $improvement, $order, $metric, $field);

	$num_criteria_met{$full_metric}++;
    }

    return $met_criteria;
}

sub find_best_output
{
    my ($outputs, $order, $metric, $field) = @_;
    my $first;
    my $best_output;
    my $best_val;

    $first = true;

    # should be checking this earlier.  we only have one run per sched
    # type per trace.
    die if scalar @$outputs != 1;

    foreach my $output (@$outputs) {
	my $val;

	$val = &get_val($output->{$metric}, $field);

	if($first) {
	    $first = false;
	    $best_output = $output;
	    $best_val = $val;
	    next;
	}

	if($order eq "high") {
	    if($val > $best_val) {
		$best_val = $val;
		$best_output = $output;
	    }
	} else {
	    if($val < $best_val) {
		$best_val = $val;
		$best_output = $output;
	    }
	}
    }

    return $best_output;
}

sub show_best_helper
{
    my ($output, $val, $label) = @_;
    my ($sched_str, $id);
    my $line;

    $line = sprintf "best %-11s: ", $label;
    $sched_str = &make_sched_str($output);
    $id = $output->{id};
    $line .= "$val (sched: $sched_str; run: $id)\n";
    &show($line);
}

sub find_improvement
{
    my ($best_baseline_val, $best_new_val, $order) = @_;
    my ($improvement, $nice_improvement);

    # on zeros...  if both values are zero, set both to the same arb
    # value so that improvement is 1.  if the denom is zero, die.

    if($best_baseline_val == 0.0 && $best_new_val == 0.0) {
	$best_baseline_val = $best_new_val = 1.0;
    }

    if($order eq "low") {

	if($best_new_val == 0.0) {
	    die "whoa: $best_baseline_val / $best_new_val\n";
	}

	$improvement = $best_baseline_val / $best_new_val;
    } else {

	if($best_baseline_val == 0.0) {
	    die "whoa: $best_new_val / $best_baseline_val\n";
	}

	$improvement = $best_new_val / $best_baseline_val;
    }

    $nice_improvement = sprintf "%.3f", $improvement;

    return $nice_improvement;
}

sub show_improvement
{
    my $improvement = shift;

    &show("best improvement: $improvement\n");
}

sub store_analysis
{
    my ($best_baseline, $best_baseline_val, $best_new, $best_new_val,
	$improvement, $order, $metric, $field) = @_;
    my ($trace_input_id, $best_baseline_id, $best_new_id);
    my $analyze_id;
    my $analyze;

    if($best_baseline->{trace_input_id} != $best_new->{trace_input_id}) {
	print "trace input ids: $best_baseline (best_baseline), " .
	    "$best_new (best_new)\n";
	die;
    }

    $trace_input_id = $best_baseline->{trace_input_id};
    $best_baseline_id = $best_baseline->{id};
    $best_new_id = $best_new->{id};

    $analyze = &analyze_data_to_hash($trace_input_id, $order, $metric, $field,
				     $best_baseline_id, $best_baseline_val,
				     $best_new_id, $best_new_val,
				     $improvement);

    # in harness, when simulating, we first check to see if the
    # simulation hasn't already been done before doing it.  and then,
    # if we do it, we lock, check to see that no one has done it
    # meanwhile, then insert.  but analysis is so fast that we don't
    # have to do the first check.  we always analyze, and then we
    # lock, check, insert.

    &db_lock_tables;
    $analyze_id = &db_find_analyze_id($trace_input_id, $order, $metric, $field,
				      $best_baseline_id, $best_new_id);
    if(defined($analyze_id)) {
	&assert_total_analyze($analyze, "current run",
			      &db_select_analyze_id($analyze_id),
			      "db id " . $analyze_id);
	# if i were to update the analyze row, i'd do it here

	if($options_quiet) {
	    $| = 1; # so single chars come out unbuffered...
	    print ".";
	    $| = 0;
	} else {
	    &show("analyze db row  : cached (id: $analyze_id, " .
		"trace: $trace_input_id)\n");
	}
    } else {
	$analyze_id = &db_insert_analyze($analyze);
	
	if($options_quiet) {
	    $| = 1; # so single chars come out unbuffered...
	    print "o";
	    $| = 0;
	} else {
	    &show("analyze db row  : stored (id: $analyze_id, " .
		"trace: $trace_input_id)\n");
	}
    }
    &db_unlock_tables;

    push @analyze_ids, $analyze_id;
}

sub analyze_data_to_hash
{
    my ($trace_input_id, $order, $metric, $field, $best_baseline_id,
	$best_baseline_val, $best_new_id, $best_new_val, $improvement) = @_;
    my %analyze;

    $analyze{trace_input_id} = $trace_input_id;
    $analyze{metric_order} = $order;
    $analyze{metric_name} = $metric;
    $analyze{metric_field} = $field;
    $analyze{baseline_id} = $best_baseline_id;
    $analyze{baseline_val} = $best_baseline_val;
    $analyze{new_id} = $best_new_id;
    $analyze{new_val} = $best_new_val;
    $analyze{improvement} = $improvement;

    return \%analyze;
}

sub assert_total_analyze
{
    my ($a, $a_key, $b, $b_key) = @_;

    &assert_helper($a->{trace_input_id},
		   $b->{trace_input_id},
		   "trace_input_id", $a_key, $b_key);
    &assert_helper($a->{metric_order},
		   $b->{metric_order},
		   "metric_order", $a_key, $b_key);
    &assert_helper($a->{metric_name},
		   $b->{metric_name},
		   "metric_name", $a_key, $b_key);
    if(defined($a->{metric_field})) {
	&assert_helper($a->{metric_field},
		       $b->{metric_field},
		       "metric_field", $a_key, $b_key);
    }
    &assert_helper($a->{baseline_id},
		   $b->{baseline_id},
		   "baseline_id", $a_key, $b_key);
    &assert_helper($a->{baseline_val},
		   $b->{baseline_val},
		   "baseline_val", $a_key, $b_key);
    &assert_helper($a->{new_id},
		   $b->{new_id},
		   "new_id", $a_key, $b_key);
    &assert_helper($a->{new_val},
		   $b->{new_val},
		   "new_val", $a_key, $b_key);
    &assert_helper($a->{improvement},
		   $b->{improvement},
		   "improvement", $a_key, $b_key);
}

# simplified from ba_harnass.pl:
sub assert_helper
{
    my ($val1, $val2, $name, $key1, $key2) = @_;
    my $good;

    $good = true;
    if($val1 =~ /[a-zA-Z]/) {
	if($val1 ne $val2) {
	    $good = false;
	}
    } else {
	if($val1 != $val2) {
	    $good = false;
	}
    }
    
    if($good == false) {
	print "val1: $val1\n";
	print "val2: $val2\n";
	die "assert outputs failure on `$name': val1 ($key1) != " .
	    "val2 ($key2)";
    }
}

sub metric_met_criteria
{
    my ($improvement, $filter_dir, $filter_amount) = @_;
    my $reportable;

    if(!defined($filter_dir)) {
	return true;
    }

    $reportable = false;
    if($filter_dir eq ">") {
	if($improvement >= $filter_amount) {
	    $reportable = true;
	}
    } elsif($filter_dir eq "<") {
	if($improvement <= $filter_amount) {
	    $reportable = true;
	}
    } else {
	die;
    }

    return $reportable;
}

sub parse_metric 
{
    my $full_metric = shift;
    my ($order,	$metric, $field, $filter_dir, $filter_amount);
    my $full_metric_without_filter;
    
    if($full_metric =~ /^([^<>]+)([<>])([\d.]+)$/) {
	$full_metric_without_filter = $1;
	$filter_dir = $2;
	$filter_amount = $3;
    } else {
	$full_metric_without_filter = $full_metric;
    }

    if($full_metric_without_filter =~ /^([^:]+):([^:]+):([^:]+)$/) {
	($order, $metric, $field) = ($1, $2, $3);
    } elsif($full_metric_without_filter =~ /^([^:]+):([^:]+)$/) {
	($order, $metric) = ($1, $2);
    } else {
	die "bad metric `$full_metric'";
    }

    return ($order, $metric, $field, $filter_dir, $filter_amount);
}

######################################################################

sub sort_numerically_callback
{
    $a <=> $b;
}

##########

sub sort_results
{
    my ($outputs_ref, $order, $metric, $field) = @_;
    my @outputs;

    undef %hack_sort_closure;
    $hack_sort_closure{order} = $order;
    $hack_sort_closure{metric} = $metric;
    $hack_sort_closure{field} = $field;
    
    @outputs = @$outputs_ref;
    @outputs = sort sort_results_callback @outputs;

    return \@outputs;
}

sub sort_results_callback
{
    my ($output_a, $output_b);
    my ($val_a, $val_b);
    my ($field, $metric, $order);

    $order = $hack_sort_closure{order};
    $metric = $hack_sort_closure{metric};
    $field = $hack_sort_closure{field};

    $output_a = $a;
    $output_b = $b;

    if(defined($field)) {
	$val_a = &get_stat_field($output_a->{$metric}, $field);
	$val_b = &get_stat_field($output_b->{$metric}, $field);
    } else {
	$val_a = $output_a->{$metric};
	$val_b = $output_b->{$metric};
    }

    if($order eq "high") {
	return $val_b <=> $val_a;
    } elsif($order eq "low") {
	return $val_a <=> $val_b;
    } else {
	die;
    }
}

##########

sub scan_sort_traces
{
    my @split_sort_traces;

    @split_sort_traces = split ",", $options_sort_traces;
    foreach my $sort_trace (@split_sort_traces) {
	my %scanned_sort_trace;
	my $direction;
	my $label;

	if($sort_trace =~ /^([+-])(\w+)$/) {
	    $direction = $1;
	    $label = $2;
	} else {
	    die "couldn't parse `$sort_trace' (forgot sort order?)";
	}
	$scanned_sort_trace{direction} = $direction;
	$scanned_sort_trace{label} = $label;
	push @scanned_sort_traces, \%scanned_sort_trace;
    }
}

sub sort_traces
{
    my $local_trace_ids = shift;
    my $local_parsed_traces = shift;

    &sort_traces_verify_option(\@scanned_sort_traces, $local_parsed_traces);

    undef %hack_sort_closure;
    $hack_sort_closure{scanned_sort_traces} = \@scanned_sort_traces;
    $hack_sort_closure{parsed_traces} = $local_parsed_traces;
    $local_trace_ids = [sort sort_traces_callback @$local_trace_ids];

    return $local_trace_ids;
}

sub sort_traces_callback
{
    my ($trace_id_a, $trace_id_b);
    my $scanned_sort_traces_ref;
    my @local_scanned_sort_traces;
    my ($parsed_trace_a, $parsed_trace_b);
    my $local_parsed_traces;

    $trace_id_a = $a;
    $trace_id_b = $b;

    $local_parsed_traces = $hack_sort_closure{parsed_traces};
    $parsed_trace_a = $local_parsed_traces->{$trace_id_a};
    $parsed_trace_b = $local_parsed_traces->{$trace_id_b};

    $scanned_sort_traces_ref = $hack_sort_closure{scanned_sort_traces};
    @local_scanned_sort_traces = @$scanned_sort_traces_ref;

    foreach my $scanned_sort_trace (@local_scanned_sort_traces) {
	my $label = $scanned_sort_trace->{label};
	my $direction = $scanned_sort_trace->{direction};
	my $val_a;
	my $val_b;

	$val_a = &sort_traces_get_val($parsed_trace_a->{$label}, $label);
	$val_b = &sort_traces_get_val($parsed_trace_b->{$label}, $label);

	if($val_a == $val_b) {
	    next;
	}

	if($direction eq "+") {
	    # need to actually parse the dists...
	    return $val_a <=> $val_b;
	} else {
	    return $val_b <=> $val_a;
	}
    }

    die "couldn't compare traces.";
}

sub sort_traces_get_val
{
    my $full_val = shift;
    my $label = shift;

    # really should be using another way...
    if($label eq "num_users" ||
       $label eq "max_req_dis_tasks_per_user" ||
       $label eq "virtual_clock_terminate") {
	return $full_val;
    } elsif($label eq "user_arrivals" ||
	    $label eq "task_set_change_range" ||
	    $label eq "tasks_per_task_set" ||
	    $label eq "service_time" ||
	    $label eq "think_time") {
	return &sort_traces_get_dist_val($full_val);
    } else {
	die "can't understand `$label'.";
    }
}

sub sort_traces_get_dist_val
{
    my $dist = shift;
    my $dist_full_val;
    my @dist_vals;
    my $num;
    my $total;

    if($dist =~ /^(.+):(.+)$/) {
	$dist_full_val = $2;
    } else {
	die "parse error";
    }

    @dist_vals = split ",", $dist_full_val;

    $num = 0;
    $total = 0;
    foreach my $dist_val (@dist_vals) {
	$num++;
	$total += $dist_val;
    }
    die "num is 0??" if $num == 0;

    return $total / $num;
}

sub sort_traces_verify_option
{
    my $scanned_sort_traces_ref = shift;
    my $local_parsed_traces = shift;
    my @local_scanned_sort_traces;
    my %count_args;
    my %unspec_args;

    @local_scanned_sort_traces = @$scanned_sort_traces_ref;

    foreach my $arg (@local_scanned_sort_traces) {
	my $label = $arg->{label};

	die "bad sort traces arg `$label'" if(!defined($trace_args{$label}));
	die "already mentioned arg `$label'" if(defined($count_args{$label}));
	$count_args{$label} = true;
    }

    foreach my $arg (keys %trace_args) {
	$unspec_args{$arg} = true;
    }
    foreach my $arg (@local_scanned_sort_traces) {
	my $label = $arg->{label};
	delete $unspec_args{$label};
    }

    foreach my $unspec_arg (keys %unspec_args) {
	my $first;
	my $comp_val;

	$first = true;
	foreach my $key (keys %$local_parsed_traces) {
	    my $parsed_trace = $local_parsed_traces->{$key};

	    if($first) {
		$first = false;
		$comp_val = $parsed_trace->{$unspec_arg};
	    } else {
		if($comp_val ne $parsed_trace->{$unspec_arg}) {
		    die "arg `$unspec_arg' changes, but isn't being sorted.";
		}
	    }
	}
    }
}

######################################################################

sub parse_options
{
    my ( %opt_hash );

    $Getopt::Long::autoabbrev = false;
    $Getopt::Long::order = $REQUIRE_ORDER;
    $Getopt::Long::ignorecase = false;

    if(GetOptions(\%opt_hash,
		  "report-metrics=s",
		  "sort-traces=s",
		  "no-sort-traces",
		  "show-settings",
		  "verbose",
		  "quiet",
		  "db-host=s",
		  "baseline-set=s",
		  "new-set=s",
		  "trace-restrict=s",
		  "req-load-hack",
		  "id-file=s") == false) {
	die "bad option(s): error in GetOptions()";
    }

    if(defined($opt_hash{"report-metrics"})) {
	@options_report_metrics = split ',', $opt_hash{"report-metrics"};
    }
    if(defined($opt_hash{"sort-traces"})) {
	$options_should_sort_traces = true;
	$options_sort_traces = $opt_hash{"sort-traces"};
    }
    if(defined($opt_hash{"no-sort-traces"})) {
	$options_should_sort_traces = false;
    }
    if(defined($opt_hash{"verbose"})) {
	$options_verbose = true;
    }
    if(defined($opt_hash{"quiet"})) {
	$options_quiet = true;
    }
    if(defined($opt_hash{"db-host"})) {
	$options_db_host = $opt_hash{"db-host"};
    }
    if(defined($opt_hash{"baseline-set"})) {
	$options_baseline_set = $opt_hash{"baseline-set"};
    }
    if(defined($opt_hash{"new-set"})) {
	$options_new_set = $opt_hash{"new-set"};
    }
    if(defined($opt_hash{"trace-restrict"})) {
	$options_trace_restrict = $opt_hash{"trace-restrict"};
    }
    if(defined($opt_hash{"req-load-hack"})) {
	$options_req_load_hack = true;
    }
    if(defined($opt_hash{"id-file"})) {
	$options_id_file = $opt_hash{"id-file"};
    }
    &show_settings if(defined($opt_hash{"show-settings"}));

    if(@ARGV >= 1) {
	foreach my $foo (@ARGV) {
	    print "$foo\n";
	}
	die "bad option(s): left over args";
    }

    die "can't turn off sorting and set a sort string"
	if(defined($opt_hash{"sort-traces"}) &&
	   defined($opt_hash{"no-sort-traces"}));
}

sub show_settings
{
    my $str;

    $str = "";
    $str .= &show_settings_helper;

    &show_force($str);
}

sub show_settings_helper
{
    my $str;

    $str = "";
    $str .= "report metrics: @options_report_metrics\n";
    if($options_should_sort_traces) {
	$str .= "sort traces: $options_sort_traces\n";
    } else {
	$str .= "sort traces: (n/a)\n";
    }
    if($options_verbose) {
	$str .= "verbose: true\n";
    } else {
	$str .= "verbose: false\n";
    }
    if($options_quiet) {
	$str .= "quiet: true\n";
    } else {
	$str .= "quiet: false\n";
    }
    $str .= "db host: $options_db_host\n";
    $str .= "baseline set: $options_baseline_set\n";
    $str .= "new set: $options_new_set\n";
    if($options_trace_restrict ne "") {
	$str .= "trace restrict: $options_trace_restrict\n";
    } else {
	$str .= "trace restrict: (n/a)\n";
    }
    if($options_req_load_hack) {
	$str .= "req_load_hack: true\n";
    } else {
	$str .= "req_load_hack: false\n";
    }
    if(defined($options_id_file)) {
	$str .= "id_file: $options_id_file\n";
    } else {
	$str .= "id_file: (none)\n";
    }

    return $str;
}

sub show_print_wrapper
{
    my $str = shift;

    print $str;
}

sub show_set_func
{
    my $new_show_func = shift;
    my $old_show_func;

    $old_show_func = $show_func;
    $show_func = $new_show_func;

    return $old_show_func;
}

sub show_buffer_func
{
    my $str = shift;

    $show_output_buffer .= $str;
}

sub show_buffer_flush
{
    return unless $show_output_buffer;

    print $show_output_buffer;
    &show_buffer_clear;
}

sub show_buffer_clear
{
    $show_output_buffer = "";
}

sub show
{
    my $str = shift;

    &show_helper($str) if($options_quiet == false);
}

sub show_verbose
{
    my $str = shift;

    &show_helper($str) if($options_verbose && $options_quiet == false);
}

sub show_force
{
    my $str = shift;

    &show_helper($str);
}

sub show_helper
{
    my $str = shift;

    &$show_func("$str");
}

######################################################################
