#include <ba_config.h>
#include <ba_sim.h>

struct ba_event_s {
    ba_id_t id;

    ba_event_kind_t kind;
    ba_real entered_abs_time;
    ba_real action_abs_time;
    ba_event_callback_t *callback;
    void *closure;
};

extern ba_id_t
ba_event_id_get(ba_event_t *event)
{
    BA_ASSERT(event);

    return event->id;
}

extern ba_event_kind_t
ba_event_kind_get(ba_event_t *event)
{
    BA_ASSERT(event);

    return event->kind;
}

extern void *
ba_event_callback_get(ba_event_t *event)
{
    BA_ASSERT(event);

    return event->callback;
}

extern void *
ba_event_closure_get(ba_event_t *event)
{
    BA_ASSERT(event);

    return event->closure;
}

extern void
ba_event_invoke_callback(ba_event_t *event)
{
    BA_ASSERT(event);
    ba_event_callback_t *callback;
    ba_event_kind_t kind;
    void *closure;
    ba_real delay_delta_time;

    callback = event->callback;
    kind = event->kind;
    closure = event->closure;
    delay_delta_time = event->action_abs_time - event->entered_abs_time;
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(delay_delta_time, 0.0);

    callback(kind, delay_delta_time, closure);
}

extern ba_real
ba_event_action_abs_time_get(ba_event_t *event)
{
    BA_ASSERT(event);

    return event->action_abs_time;
}

extern ba_real
ba_event_entered_abs_time_get(ba_event_t *event)
{
    BA_ASSERT(event);

    return event->entered_abs_time;
}

extern void
ba_event_print(ba_event_t *event)
{
    BA_ASSERT(event);

    ba_print_info("event: %p\n", event);
    ba_print_info("  id: " BA_ID_FMT "\n", event->id);
    switch(event->kind) {
    case BA_EVENT_KIND_RESOURCE_IDLE:
	ba_print_info("  event kind: resource idle\n");
	break;
    case BA_EVENT_KIND_RESOURCE_BUSY:
	ba_print_info("  event kind: resource busy\n");
	break;
    case BA_EVENT_KIND_USER_REQUEST_TASK:
	ba_print_info("  event kind: user request task\n");
	break;
    case BA_EVENT_KIND_USER_ACTIVATE_NEW_TASK_SET:
	ba_print_info("  event kind: user activate new task set\n");
	break;
    case BA_EVENT_KIND_USERS_ADD_USER:
	ba_print_info("  event kind: users add user\n");
	break;
    case BA_EVENT_KIND_USERS_REMOVE_USER:
	ba_print_info("  event kind: users remove user\n");
	break;
    case BA_EVENT_KIND_USERS_VC_TERMINATE:
	ba_print_info("  event kind: users virtual clock terminate\n");
	break;
    case BA_EVENT_KIND_UNKNOWN:
	BA_PANIC();
	break;
    }
    ba_print_info("  entered abs time: " BA_REAL_FMT "\n",
		  event->entered_abs_time);
    ba_print_info("  action abs time: " BA_REAL_FMT "\n",
		  event->action_abs_time);
    ba_print_info("  callback: %p\n", event->callback);
    ba_print_info("  closure: %p\n", event->closure);
}

extern void
ba_event_print_from_list(void *event_in,
			 void *unused)
{
    BA_ASSERT(event_in);
    ba_event_t *event;
    
    event = (ba_event_t *)event_in;

    ba_event_print(event);
}

extern ba_event_t *
ba_event_new(ba_id_t id,
	     ba_event_kind_t kind,
	     ba_real entered_abs_time,
	     ba_real action_abs_time,
	     ba_event_callback_t callback,
	     void *closure)
{
    BA_ASSERT(callback);
    ba_event_t *event;

    event = ba_mem_malloc(sizeof(ba_event_t));

    event->id = id;

    event->kind = kind;
    event->entered_abs_time = entered_abs_time;
    event->action_abs_time = action_abs_time;
    event->callback = callback;
    event->closure = closure;

    return event;
}

extern void
ba_event_delete(ba_event_t *event)
{
    BA_ASSERT(event);

    ba_mem_free(event);
}

extern void
ba_event_delete_from_list(void *event_in,
			  void *unused)
{
    BA_ASSERT(event_in);
    ba_event_t *event;

    event = (ba_event_t *)event_in;

    ba_event_delete(event);
}
