#include <ba_config.h>
#include <string.h>
#include <ba_sim.h>

struct ba_glib_fast_slist_append_s {
    GSList *appender;
};

// not using ba_mem_malloc or ba_mem_free here...  just in case that `class'
// hasn't been init'd when they need to be called from here.

static void glib_print_handler(const gchar *s);
static void glib_printerr_handler(const gchar *s);

static void check_for_leaks(void);
static void hack_to_read_leak_info(const gchar *s);

static GAllocator *slist_allocator;
static GPrintFunc old_print_handler;
static GPrintFunc old_printerr_handler;

static ba_bool we_leaked;

extern GHookList *
ba_glib_hook_new(void)
{
    GHookList *list;

    list = g_malloc(sizeof(GHookList));
    g_hook_list_init(list, sizeof(GHook));

    return list;
}

extern void
ba_glib_hook_delete(GHookList *list)
{
    BA_ASSERT(list);

    g_hook_list_clear(list);
    g_free(list);
}

extern void
ba_glib_hook_add(GHookList *list,
		 GHookFunc func)
{
    BA_ASSERT(list && func);
    GHook *hook;

    hook = g_hook_alloc(list);
    hook->func = func;
    g_hook_prepend(list, hook);
}

extern GSList *
ba_glib_fast_slist_append(ba_glib_fast_slist_append_t *fast_slist_append,
			  GSList *head,
			  gpointer data)
{
    BA_ASSERT(fast_slist_append);

    fast_slist_append->appender = g_slist_append(fast_slist_append->appender,
						 data);
    if(head == 0) {
	head = fast_slist_append->appender;
    } else {
	fast_slist_append->appender =
	    g_slist_next(fast_slist_append->appender);
    }

    return head;
}

extern void
ba_glib_fast_slist_appender_reset(ba_glib_fast_slist_append_t
				  *fast_slist_append)
{
    BA_ASSERT(fast_slist_append);

    fast_slist_append->appender = 0;
}

extern ba_glib_fast_slist_append_t *
ba_glib_fast_slist_append_new(void)
{
    ba_glib_fast_slist_append_t *fast_slist_append;

    fast_slist_append = g_malloc(sizeof(ba_glib_fast_slist_append_t));
    fast_slist_append->appender = 0;

    return fast_slist_append;
}

extern void
ba_glib_fast_slist_append_delete(ba_glib_fast_slist_append_t
				 *fast_slist_append)
{
    BA_ASSERT(fast_slist_append);

    // nothing to do...
    g_free(fast_slist_append);
}

extern ba_status
ba_glib_init(void)
{
    we_leaked = BA_FALSE;

    g_mem_set_vtable(glib_mem_profiler_table);

    slist_allocator = g_allocator_new("BA g_slist allocator", 128);
    g_slist_push_allocator(slist_allocator);

    // actually, so long as i don't call g_print() or g_printerr(),
    // the following two handlers are not strictly necessary.
    old_print_handler = g_set_print_handler(glib_print_handler);
    old_printerr_handler = g_set_printerr_handler(glib_printerr_handler);

    return BA_STATUS_SUCCESS;
}

static void
glib_print_handler(const gchar *s)
{
    ba_print_info(s);
}

static void
glib_printerr_handler(const gchar *s)
{
    ba_print_info(s);
}

extern ba_status
ba_glib_destroy(void)
{
    g_slist_pop_allocator();
    g_allocator_free(slist_allocator);

    check_for_leaks();

    g_set_print_handler(old_print_handler);
    g_set_printerr_handler(old_printerr_handler);

    return BA_STATUS_SUCCESS;
}

static void
check_for_leaks(void)
{
    static GPrintFunc tmp_print_handler;

    tmp_print_handler = g_set_print_handler(hack_to_read_leak_info);
    g_mem_profile();
    g_set_print_handler(tmp_print_handler);

    if(we_leaked) {
	ba_print_error("A memory leak has occurred:\n");
	tmp_print_handler = g_set_print_handler(glib_printerr_handler);
	g_mem_profile();
	g_set_print_handler(tmp_print_handler);
	BA_PANIC();
    }
}

// XXX probably will fail on a glib different than 2.2.1.
static void
hack_to_read_leak_info(const gchar *s)
{
    uint32_t len;
    char *cmp = "Total bytes: allocated";

    if(strncmp(cmp, s, strlen(cmp)) != 0) {
	return;
    }

    len = strlen(s);
    if(s[len - 3] != '=' ||
       s[len - 2] != '0') {
	we_leaked = BA_TRUE;
    }
}
