#include <ba_config.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <ba_sim.h>

static ba_bool quiet = BA_FALSE;

static FILE *info_file = 0;
static FILE *data_file = 0;
static FILE *error_file = 0;

static FILE *trace_output_file = 0;
static FILE *trace_input_file = 0;

static ba_bool get_str_from_trace_input(char *str,
					uint32_t str_len);
static ba_bool omit_rest_of_line_from_trace_input(void);

extern void
ba_print_info(const char *fmt, ...)
{
    va_list ap;
    BA_ASSERT(fmt); // keep this assert under 'ap' just to be sure
    FILE *file = info_file ? info_file : stdout;

    if(quiet == BA_TRUE) {
	va_start(ap, fmt);
	va_end(ap);
	return;
    }

    va_start(ap, fmt);
    vfprintf(file, fmt, ap);
    va_end(ap);

    ba_print_flush_info(); // XXX for debugging purposes
}

extern void
ba_print_data(const char *fmt, ...)
{
    va_list ap;
    BA_ASSERT(fmt); // keep this assert under 'ap' just to be sure
    FILE *file = data_file ? data_file : stdout;

    va_start(ap, fmt);
    vfprintf(file, fmt, ap);
    va_end(ap);
}

extern void
ba_print_error(const char *fmt, ...)
{
    va_list ap;
    BA_ASSERT(fmt); // keep this assert under 'ap' just to be sure
    FILE *file = error_file ? error_file : stderr;

    va_start(ap, fmt);
    vfprintf(file, fmt, ap);
    va_end(ap);
}

extern void
ba_print_write_trace_output(const char *fmt, ...)
{
    va_list ap;
    BA_ASSERT(fmt); // keep this assert under 'ap' just to be sure

    BA_ASSERT(trace_output_file);
    va_start(ap, fmt);
    vfprintf(trace_output_file, fmt, ap);
    va_end(ap);
}

extern ba_bool
ba_print_read_trace_input_line(char *str,
			       uint32_t str_len,
			       ba_bool omit_extra)
{
    if(get_str_from_trace_input(str, str_len)) {
	return BA_TRUE;
    }

    if(omit_extra) {
	if(strlen(str) == str_len) {
	    return omit_rest_of_line_from_trace_input();
	} else {
	    return BA_FALSE;
	}
    } else {
	// if i filled up to str_len chars, i _know_ something was wrong.
	BA_ASSERT(strlen(str) < str_len);
	return BA_FALSE;
    }
}

static ba_bool
get_str_from_trace_input(char *str,
			 uint32_t str_len)
{
    char *err;

    err = fgets(str, str_len + 1, trace_input_file);
    BA_ASSERT(strlen(str) <= str_len);
    if(err == 0) {
	BA_ASSERT(ferror(trace_input_file) == 0);
	return BA_TRUE;
    } else {
	return BA_FALSE;
    }
}

static ba_bool
omit_rest_of_line_from_trace_input(void)
{
    char *omit_str;

    omit_str = ba_string_new_len("", 80);
    while(1) {
	if(get_str_from_trace_input(omit_str, 80)) {
	    ba_string_delete(omit_str);
	    return BA_TRUE;
	}
	if(strlen(omit_str) < 80) {
	    break;
	}
    }

    ba_string_delete(omit_str);
    return BA_FALSE;
}

extern void
ba_print_flush_info(void)
{
    FILE *file = info_file ? info_file : stdout;
    int err;

    err = fflush(file);
    if(err != 0) {
	if(errno == EPIPE) {
	    return;
	}
    }
    BA_ASSERT(err == 0);
}

extern void
ba_print_flush_data(void)
{
    FILE *file = data_file ? data_file : stdout;
    int err;

    err = fflush(file);
    if(err != 0) {
	if(errno == EPIPE) {
	    return;
	}
    }
    BA_ASSERT(err == 0);
}

extern void
ba_print_flush_error(void)
{
    FILE *file = error_file ? error_file : stderr;
    int err;

    err = fflush(file);
    if(err != 0) {
	if(errno == EPIPE) {
	    return;
	}
    }
    BA_ASSERT(err == 0);
}

extern void
ba_print_flush_trace_output(void)
{
    int err;

    if(trace_output_file == 0) {
	return;
    }

    err = fflush(trace_output_file);
    if(err != 0) {
	if(errno == EPIPE) {
	    return;
	}
    }
    BA_ASSERT(err == 0);
}

extern ba_status
ba_print_init(void)
{
    char *data_output_filename;
    char *trace_output_filename;
    char *trace_input_filename;

    quiet = ba_options_quiet_get();
    data_output_filename = ba_options_data_output_filename_get();

    trace_output_filename = ba_options_trace_output_filename_get();
    trace_input_filename = ba_options_trace_input_filename_get();

    info_file = stdout;

    if(data_output_filename) {
	data_file = fopen(data_output_filename, "w");
	if(data_file == 0) {
	    // yes, it's ok to call ba_print_error() from this init,
	    // since i do something smart if called before init'ing...
	    ba_print_error("Error trying to open `%s' for data output "
			   "writing: %s (%d)\n", data_output_filename,
			   strerror(errno), errno);
	    return BA_STATUS_PRINT_BAD_DATA_OUTPUT_FILENAME;
	}
    } else {
	data_file = stdout;
    }

    if(trace_output_filename) {
	trace_output_file = fopen(trace_output_filename, "w");
	if(trace_output_file == 0) {
	    ba_print_error("Error trying to open `%s' for trace output "
			   "writing: %s (%d)\n", trace_output_filename,
			   strerror(errno), errno);
	    return BA_STATUS_PRINT_BAD_TRACE_OUTPUT_FILENAME;
	}
    }

    if(trace_input_filename) {
	trace_input_file = fopen(trace_input_filename, "r");
	if(trace_input_file == 0) {
	    ba_print_error("Error trying to open `%s' for trace input "
			   "writing: %s (%d)\n", trace_input_filename,
			   strerror(errno), errno);
	    return BA_STATUS_PRINT_BAD_TRACE_INPUT_FILENAME;
	}
    }

    error_file = stderr;

    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_print_destroy(void)
{
    int err;

    ba_print_flush_info();
    ba_print_flush_data();
    ba_print_flush_error();

    ba_print_flush_trace_output();

    if(data_file != stdout) {
	err = fclose(data_file);
	BA_ASSERT(err == 0);
    }

    if(trace_output_file) {
	err = fclose(trace_output_file);
	BA_ASSERT(err == 0);
    }

    if(trace_input_file) {
	err = fclose(trace_input_file);
	BA_ASSERT(err == 0);
    }

    return BA_STATUS_SUCCESS;
}

extern void
ba_print_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_print_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
