#include <ba_config.h>
#include <ba_sim.h>
#include <time.h>

#define SAVE_MEMORY 1 // XXX ./configure option?

typedef struct {
    // scheduler
    ba_time_period_t *virtual_clock_time_period;
    ba_real virtual_clock_delta_time;
    ba_stats_running_mean_t *all_candidates_rm_stats;
    ba_stats_running_mean_t *req_candidates_rm_stats;
    ba_stats_running_mean_t *dis_candidates_rm_stats;
    ba_stats_running_mean_t *weighed_all_candidates_rm_stats;
    ba_stats_running_mean_t *weighed_req_candidates_rm_stats;
    ba_stats_running_mean_t *weighed_dis_candidates_rm_stats;

    // resources
    ba_stats_t *resource_usage_stats;
    ba_stats_t *req_resource_usage_stats;
    ba_stats_t *dis_resource_usage_stats;
    ba_stats_t *scaled_resource_usage_stats;
    ba_stats_t *billable_resources_usage_stats; // pointer to one of the above
    ba_time_periods_t *resource_busy_periods;
    ba_real resource_busy_first_abs_time;
    ba_real resource_busy_last_abs_time;
    ba_real resource_busy_range_delta_time;
    ba_real resource_busy_delta_time;
    ba_real resource_idle_delta_time;
    ba_time_periods_t *non_work_conserving_periods;
    ba_real non_work_conserving_delta_time;
    ba_real load;
    ba_real req_load;
    ba_real req_load_ignoring_idle_delta_time;

    // tasks and task sets
    uint32_t user_count;
    uint32_t num_hit_req_dis_limit;
    ba_stats_t *task_sets_stats;
    ba_time_periods_t *task_set_alive_periods;
    ba_real task_set_alive_delta_time;
    ba_time_periods_t *no_task_sets_periods;
    ba_real no_task_sets_delta_time;
    // from ba_task_set_report, we see that the following var tracks
    // (1) when the task_set was created to when the last task
    // executed (`last' as in the last, final task in the task set,
    // not as in the most recently executed task) or (2) the entire
    // length of the task_set (equal to task_set_alive_periods?) if nothing
    // executed.  task_set_ready_tasks_periods could be smaller than
    // task_set_alive_periods because a task_set could have all its
    // tasks executed while a user is still thinking through the
    // output of other tasks.
    ba_time_periods_t *task_set_ready_tasks_periods;
    ba_real task_set_ready_tasks_delta_time;
    ba_stats_t *disclosed_tasks_stats;
    ba_stats_t *disclosed_only_tasks_stats;
    ba_stats_t *executed_tasks_stats;
    ba_stats_t *requested_tasks_stats;
    ba_stats_t *finished_tasks_stats;
    ba_stats_t *canceled_tasks_stats;
    ba_stats_t *service_delta_time_stats;
    ba_stats_t *think_delta_time_stats;

    // visible response time
    ba_stats_t *vrt_delta_time_stats;
    ba_stats_t *vsd_stats;
    ba_stats_t *user_vrt_delta_time_stats;
    ba_stats_t *user_vsd_stats;
    ba_stats_t *deadlines_met_stats;
} report_t;

// 2 days, XXX should be an option
#if BA_SIM_INSPECT
static ba_real warmup_abs_time = 0.0;
#else
static ba_real warmup_abs_time = 60 * 60 * 24 * 2;
#endif

static report_t *report;

static void process_report(void);

static void print_proem(void);
static void print_scheduler_params(void);
static void print_user_behavior(void);
static void print_user_traced_behavior_to(ba_print_func_t *print_func);
static void print_user_parameter_behavior_to(ba_print_func_t *print_func);
static void print_scheduler(void);
static void print_resources(void);
static void print_tasks(void);
static void print_vrt(void);

static void assert_scheduler(void);
static void assert_resources(void);
static void assert_tasks(void);
static void assert_vrt(void);

// scheduler

extern void
ba_report_events_virtual_clock_start_end(ba_real start_abs_time,
					 ba_real end_abs_time)
{
    report->virtual_clock_time_period = ba_time_period_new(start_abs_time,
							   end_abs_time);
}

extern void
ba_report_sched_candidates(uint32_t num_all_candidates,
			   uint32_t num_req_candidates,
			   uint32_t num_dis_candidates,
			   ba_real abs_time)
{
    ba_stats_running_mean_insert(report->all_candidates_rm_stats,
				 num_all_candidates, abs_time);
    ba_stats_running_mean_insert(report->req_candidates_rm_stats,
				 num_req_candidates, abs_time);
    ba_stats_running_mean_insert(report->dis_candidates_rm_stats,
				 num_dis_candidates, abs_time);
}

extern void
ba_report_sched_weighed_candidates(uint32_t weighed_all_candidates,
				   uint32_t weighed_req_candidates,
				   uint32_t weighed_dis_candidates,
				   ba_real abs_time)
{
    ba_stats_running_mean_insert(report->weighed_all_candidates_rm_stats,
				 weighed_all_candidates, abs_time);
    ba_stats_running_mean_insert(report->weighed_req_candidates_rm_stats,
				 weighed_req_candidates, abs_time);
    ba_stats_running_mean_insert(report->weighed_dis_candidates_rm_stats,
				 weighed_dis_candidates, abs_time);
}

// resources

extern void
ba_report_user_resource_usage(ba_real resource_usage)
{
    ba_stats_insert(report->resource_usage_stats, resource_usage);
}

extern void
ba_report_user_req_resource_usage(ba_real req_resource_usage)
{
    ba_stats_insert(report->req_resource_usage_stats, req_resource_usage);
}

extern void
ba_report_user_dis_resource_usage(ba_real dis_resource_usage)
{
    ba_stats_insert(report->dis_resource_usage_stats, dis_resource_usage);
}

extern void
ba_report_user_scaled_resource_usage(ba_real scaled_resource_usage)
{
    ba_stats_insert(report->scaled_resource_usage_stats, scaled_resource_usage);
}

extern void
ba_report_resource_busy_start_end(ba_real start_abs_time,
				  ba_real end_abs_time)
{
    ba_time_periods_append_start_end(report->resource_busy_periods,
				     start_abs_time, end_abs_time);
}

// tasks and task sets

extern void
ba_report_users_inc_user_count(void)
{
    report->user_count++;
}

extern void
ba_report_user_inc_num_hit_req_dis_limit(void)
{
    report->num_hit_req_dis_limit++;
}

extern void
ba_report_user_task_sets(uint32_t task_sets)
{
    ba_stats_insert(report->task_sets_stats, task_sets);
}

extern void
ba_report_task_set_alive_start_end(ba_real start_abs_time,
				   ba_real end_abs_time)
{
#if !SAVE_MEMORY
    ba_time_periods_union_start_end(report->task_set_alive_periods,
				    start_abs_time, end_abs_time,
				    report->task_set_alive_periods);
#endif
}

extern void
ba_report_task_set_ready_tasks_start_end(ba_real start_abs_time,
					 ba_real end_abs_time)
{
#if !SAVE_MEMORY
    ba_time_periods_union_start_end(report->task_set_ready_tasks_periods,
				    start_abs_time, end_abs_time,
				    report->task_set_ready_tasks_periods);
#endif
}

extern void
ba_report_user_disclosed_tasks(uint32_t num_tasks)
{
    ba_stats_insert(report->disclosed_tasks_stats, num_tasks);
}

extern void
ba_report_user_disclosed_only_tasks(uint32_t num_tasks)
{
    ba_stats_insert(report->disclosed_only_tasks_stats, num_tasks);
}

extern void
ba_report_user_executed_tasks(uint32_t num_tasks)
{
    ba_stats_insert(report->executed_tasks_stats, num_tasks);
}

extern void
ba_report_user_requested_tasks(uint32_t num_tasks)
{
    ba_stats_insert(report->requested_tasks_stats, num_tasks);
}

extern void
ba_report_user_finished_tasks(uint32_t num_tasks)
{
    ba_stats_insert(report->finished_tasks_stats, num_tasks);
}

extern void
ba_report_user_canceled_tasks(uint32_t num_tasks)
{
    ba_stats_insert(report->canceled_tasks_stats, num_tasks);
}

extern void
ba_report_user_service_delta_time(ba_real service_delta_time)
{
    ba_stats_insert(report->service_delta_time_stats, service_delta_time);
}

extern void
ba_report_user_think_delta_time(ba_real think_delta_time)
{
    ba_stats_insert(report->think_delta_time_stats, think_delta_time);
}

// visible response time

extern void
ba_report_user_vrt_delta_time(ba_real vrt_delta_time)
{
    ba_stats_insert(report->vrt_delta_time_stats, vrt_delta_time);
}

extern void
ba_report_user_vsd(ba_real vsd)
{
    ba_stats_insert(report->vsd_stats, vsd);
}

extern void
ba_report_user_user_vrt_delta_time(ba_real user_vrt_delta_time)
{
    ba_stats_insert(report->user_vrt_delta_time_stats,
		    user_vrt_delta_time);
}

extern void
ba_report_user_user_vsd(ba_real user_vsd)
{
    ba_stats_insert(report->user_vsd_stats, user_vsd);
}

extern void
ba_report_user_deadlines_met(uint32_t deadlines_met)
{
    ba_stats_insert(report->deadlines_met_stats, deadlines_met);
}

//////////////////////////////////////////////////////////////////////

static void
process_report(void)
{
    // scheduler
    report->virtual_clock_delta_time =
	ba_time_period_delta_time(report->virtual_clock_time_period);
    
    // resources
    if(ba_options_treat_req_dis_diff_get()) {
	// inter or ba sched:
	report->billable_resources_usage_stats =
	    report->req_resource_usage_stats;
    } else {
	// batch sched:
	report->billable_resources_usage_stats = report->resource_usage_stats;
    }

    report->resource_busy_first_abs_time = 
	ba_time_periods_first_start_abs_time(report->resource_busy_periods);
    report->resource_busy_last_abs_time =
	ba_time_periods_last_end_abs_time(report->resource_busy_periods);
    report->resource_busy_range_delta_time =
	report->resource_busy_last_abs_time -
	report->resource_busy_first_abs_time;
    report->resource_busy_delta_time =
	ba_time_periods_total_delta_time(report->resource_busy_periods);
    report->resource_idle_delta_time = report->resource_busy_range_delta_time -
	report->resource_busy_delta_time;

#if !SAVE_MEMORY
    ba_time_periods_difference(report->task_set_ready_tasks_periods,
			       report->resource_busy_periods,
			       report->non_work_conserving_periods);
#endif

    report->non_work_conserving_delta_time =
	ba_time_periods_total_delta_time(report->non_work_conserving_periods);

    if(ba_options_virtual_clock_terminate_get()) {
	if(ba_types_real_equal(ba_options_virtual_clock_terminate_get() -
			       ba_report_warmup_abs_time_get(), 0.0)) {
	    BA_PANIC();
	    report->load = 0.0;
	    report->req_load = 0.0;
	} else {
	    report->load = ba_stats_total_get(report->resource_usage_stats) /
		(ba_options_virtual_clock_terminate_get() -
		 ba_report_warmup_abs_time_get());
	    report->req_load = ba_stats_total_get(report->
						  req_resource_usage_stats) /
		(ba_options_virtual_clock_terminate_get() -
		 ba_report_warmup_abs_time_get());
	}
    } else {
	// this `no virtual clock terminate' case may be broke XXX
	if(report->resource_busy_range_delta_time == 0) {
	    report->load = 0.0;
	    report->req_load = 0.0;
	} else {
	    report->load = report->resource_busy_delta_time /
		report->resource_busy_range_delta_time;
	    report->req_load =
		ba_stats_total_get(report->req_resource_usage_stats) /
		report->resource_busy_range_delta_time;
	}
    }

    if(ba_stats_total_get(report->resource_usage_stats) == 0.0) {
	report->req_load_ignoring_idle_delta_time = 0.0;
    } else {
	report->req_load_ignoring_idle_delta_time =
	    ba_stats_total_get(report->req_resource_usage_stats) /
	    ba_stats_total_get(report->resource_usage_stats);
    }

    // tasks and task sets
    report->task_set_alive_delta_time =
	ba_time_periods_total_delta_time(report->task_set_alive_periods);

#if !SAVE_MEMORY
    ba_time_periods_difference_period_r(report->virtual_clock_time_period,
					report->task_set_alive_periods,
					report->no_task_sets_periods);
#endif

    report->no_task_sets_delta_time =
	ba_time_periods_total_delta_time(report->no_task_sets_periods);
    report->task_set_ready_tasks_delta_time =
	ba_time_periods_total_delta_time(report->task_set_ready_tasks_periods);

    // visible response time
    // (nothing to process)
}

//////////////////////////////////////////////////////////////////////

static void
print_proem(void)
{
    ba_report_print_proem_to(ba_print_data);
}

extern void
ba_report_print_proem_to(ba_print_func_t *print_func)
{
    time_t time_struct;

    print_func("---- proem ----\n");
    print_func("program: " PACKAGE_NAME " " PACKAGE_VERSION "\n");
    print_func("args: %s\n", ba_options_command_line_options_get());
    // actually check ret values of all this time stuff.  and do i
    // want it to be from the time simulation started or ended?  in a
    // trace, the following will be at the start, and in normal
    // output, the following time will be at the end. XXX
    time(&time_struct);
    print_func("run date: UTC %s", asctime(gmtime(&time_struct)));
    print_func("random seed: %u\n", ba_options_random_seed_get());
    print_func("\n");
}

static void
print_scheduler_params(void)
{
    ba_print_data("---- scheduler parameters ----\n");

    ba_print_data("warmup time (for selected metrics): "
		  BA_REAL_LOW_PREC_FMT "\n", warmup_abs_time);
    ba_print_data("treat requested and disclosed tasks differently: ");
    ba_types_bool_print_to(ba_options_treat_req_dis_diff_get(), ba_print_data);
    ba_print_data("\n");
    ba_print_data("requested scheduler behavior: ");
    ba_sched_behavior_print_short_to(ba_options_req_sched_behavior_get(),
				     ba_print_data);
    ba_print_data("\n");
    ba_print_data("disclosed scheduler behavior: ");
    if(ba_options_treat_req_dis_diff_get()) {
	ba_sched_behavior_print_short_to(ba_options_dis_sched_behavior_get(),
					 ba_print_data);
    } else {
	ba_print_data("n/a (disclosed tasks treated as requested)");
    }
    ba_print_data("\n");
    ba_print_data("\n");
}

static void
print_user_behavior(void)
{
    ba_report_print_user_behavior_to(ba_print_data);
}

extern void
ba_report_print_user_behavior_to(ba_print_func_t *print_func)
{
    print_func("---- user behavior ----\n");
    if(ba_trace_read_trace_input_get()) {
	print_user_traced_behavior_to(print_func);
    } else {
	print_user_parameter_behavior_to(print_func);
    }
}

static void
print_user_traced_behavior_to(ba_print_func_t *print_func)
{
    print_func("trace input: %s\n", ba_options_trace_input_filename_get());
    BA_ASSERT(ba_trace_write_trace_output_get() == BA_FALSE);
    print_func("trace output: (none)\n");
    print_func("\n");
}

static void
print_user_parameter_behavior_to(ba_print_func_t *print_func)
{
    BA_ASSERT(ba_trace_read_trace_input_get() == BA_FALSE);
    print_func("trace input: (none)\n");
    if(ba_trace_write_trace_output_get()) {
	print_func("trace output: %s\n",
		   ba_options_trace_output_filename_get());
    } else {
	print_func("trace output: (none)\n");
    }
    print_func("number of users: %u\n", ba_options_num_users_get());
    print_func("virtual clock terminate: " BA_REAL_LOW_PREC_FMT "\n",
	       ba_options_virtual_clock_terminate_get());
    print_func("maximum requested and disclosed only tasks per user: "
	       "%u\n", ba_options_max_req_dis_tasks_per_user_get());
    print_func("user arrivals: ");
    ba_rv_dist_print_short_to(ba_options_user_arrivals_dist_get(),
			      print_func);
    print_func("\n");
    print_func("task set change range: ");
    ba_rv_dist_print_short_to(ba_options_task_set_change_range_dist_get(),
			      print_func);
    print_func("\n");
    print_func("tasks per task set: ");
    ba_rv_dist_print_short_to(ba_options_tasks_per_task_set_dist_get(),
			      print_func);
    print_func("\n");
    print_func("service time: ");
    ba_rv_dist_print_short_to(ba_options_service_delta_time_dist_get(),
			      print_func);
    print_func("\n");
    print_func("think time: ");
    ba_rv_dist_print_short_to(ba_options_think_delta_time_dist_get(),
			      print_func);
    print_func("\n");
    print_func("\n");
}

static void
print_scheduler(void)
{
    ba_print_data("---- scheduler ----\n");
    ba_print_data("virtual clock: " BA_REAL_LOW_PREC_FMT "\n",
		  report->virtual_clock_delta_time);
    ba_print_data("all candidates queue: " BA_REAL_LOW_PREC_FMT ", "
		  BA_REAL_LOW_PREC_FMT "\n",
		  ba_stats_running_mean_get(report->all_candidates_rm_stats),
		  ba_stats_running_mean_get(report->
					    weighed_all_candidates_rm_stats));
    ba_print_data("requested candidates queue: " BA_REAL_LOW_PREC_FMT ", "
		  BA_REAL_LOW_PREC_FMT "\n",
		  ba_stats_running_mean_get(report->req_candidates_rm_stats),
		  ba_stats_running_mean_get(report->
					    weighed_req_candidates_rm_stats));
    ba_print_data("disclosed candidates queue: " BA_REAL_LOW_PREC_FMT ", "
		  BA_REAL_LOW_PREC_FMT "\n",
		  ba_stats_running_mean_get(report->dis_candidates_rm_stats),
		  ba_stats_running_mean_get(report->
					    weighed_dis_candidates_rm_stats));
    ba_print_data("\n");
}

static void
print_resources(void)
{
    ba_print_data("---- resources ----\n");
    ba_print_data("all resource usage: ");
    ba_stats_print_nice_to(report->resource_usage_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("requested resource usage: ");
    ba_stats_print_nice_to(report->req_resource_usage_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("disclosed only rsrc usage: ");
    ba_stats_print_nice_to(report->dis_resource_usage_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("scaled resource usage (all / req.): ");
    ba_stats_print_nice_to(report->scaled_resource_usage_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("billable resources: ");
    ba_stats_print_nice_to(report->billable_resources_usage_stats,
			   ba_print_data);
    ba_print_data("\n");
    ba_print_data("resource busy range time: " BA_REAL_LOW_PREC_FMT "\n",
		  report->resource_busy_range_delta_time);
    ba_print_data("resource busy time: " BA_REAL_LOW_PREC_FMT
		  " (resource idle time: " BA_REAL_LOW_PREC_FMT ")\n",
		  report->resource_busy_delta_time,
		  report->resource_idle_delta_time);
    ba_print_data("non-work conserving time: " BA_REAL_LOW_PREC_FMT "\n",
		  report->non_work_conserving_delta_time);
    ba_print_data("load: " BA_REAL_FMT "\n", report->load);
    ba_print_data("requested load: " BA_REAL_FMT "\n", report->req_load);
    ba_print_data("requested load ignoring idle time: " BA_REAL_FMT "\n",
		  report->req_load_ignoring_idle_delta_time);
    ba_print_data("\n");
}

static void
print_tasks(void)
{
    ba_print_data("---- tasks and task sets ----\n");
    ba_print_data("user count: %d\n", report->user_count);
    ba_print_data("times the req. & dis. max was reached: %d\n",
		  report->num_hit_req_dis_limit);
    ba_print_data("task sets: ");
    ba_stats_print_nice_to(report->task_sets_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("task sets alive time: " BA_REAL_LOW_PREC_FMT
		  " (no task sets alive time: " BA_REAL_LOW_PREC_FMT ")\n",
		  report->task_set_alive_delta_time,
		  report->no_task_sets_delta_time);
    ba_print_data("tasks ready time: " BA_REAL_LOW_PREC_FMT "\n",
		  report->task_set_ready_tasks_delta_time);
    ba_print_data("disclosed tasks: ");
    ba_stats_print_nice_to(report->disclosed_tasks_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("disclosed only tasks: ");
    ba_stats_print_nice_to(report->disclosed_only_tasks_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("executed tasks: ");
    ba_stats_print_nice_to(report->executed_tasks_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("requested tasks: ");
    ba_stats_print_nice_to(report->requested_tasks_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("finished tasks: ");
    ba_stats_print_nice_to(report->finished_tasks_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("canceled tasks: ");
    ba_stats_print_nice_to(report->canceled_tasks_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("\n");
}

static void
print_vrt(void)
{
    ba_print_data("---- visible response time ----\n");
    ba_print_data("visible response time: ");
    ba_stats_print_nice_maybe_none_to(report->vrt_delta_time_stats,
				      ba_print_data);
    ba_print_data("\n");
    ba_print_data("visible slowdown (vrt / st): ");
    ba_stats_print_nice_maybe_none_to(report->vsd_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("u. visible response time: ");
    ba_stats_print_nice_to(report->user_vrt_delta_time_stats,
			   ba_print_data);
    ba_print_data("\n");
    ba_print_data("u. visible slowdown: ");
    ba_stats_print_nice_to(report->user_vsd_stats, ba_print_data);
    ba_print_data("\n");
    ba_print_data("deadlines met: ");
    ba_stats_print_nice_to(report->deadlines_met_stats, ba_print_data);
    ba_print_data("\n");
}

//////////////////////////////////////////////////////////////////////

static
void assert_scheduler(void)
{
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->virtual_clock_delta_time,
					 0.0);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->virtual_clock_delta_time,
					 report->resource_busy_delta_time);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->virtual_clock_delta_time,
					 report->resource_idle_delta_time);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->virtual_clock_delta_time,
					 report->task_set_alive_delta_time);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->virtual_clock_delta_time,
					 report->
					 task_set_ready_tasks_delta_time);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->virtual_clock_delta_time,
					 report->no_task_sets_delta_time);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->virtual_clock_delta_time,
					 report->
					 non_work_conserving_delta_time);
    
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->all_candidates_rm_stats), 0.0);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->all_candidates_rm_stats),
      ba_stats_running_mean_get(report->req_candidates_rm_stats));
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->all_candidates_rm_stats),
      ba_stats_running_mean_get(report->dis_candidates_rm_stats));

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->req_candidates_rm_stats), 0.0);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->dis_candidates_rm_stats), 0.0);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->weighed_all_candidates_rm_stats),
      0.0);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->weighed_all_candidates_rm_stats),
      ba_stats_running_mean_get(report->weighed_req_candidates_rm_stats));
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->weighed_all_candidates_rm_stats),
      ba_stats_running_mean_get(report->weighed_dis_candidates_rm_stats));

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->weighed_req_candidates_rm_stats),
      0.0);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_running_mean_get(report->weighed_dis_candidates_rm_stats),
      0.0);
}

static
void assert_resources(void)
{
    uint32_t max_req_dis_tasks_per_user;
    uint32_t user_count;

    max_req_dis_tasks_per_user =
	  ba_behavior_max_req_dis_tasks_per_user_get();

    user_count = ba_behavior_num_users_get();

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->resource_usage_stats), 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
      ba_stats_max_get(report->resource_usage_stats),
      report->resource_busy_delta_time);
    BA_ASSERT(ba_stats_count_get(report->resource_usage_stats) ==
	      user_count);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_total_get(report->resource_usage_stats), 0.0);
    BA_ASSERT_REAL_EQUAL(ba_stats_total_get(report->req_resource_usage_stats) +
			 ba_stats_total_get(report->dis_resource_usage_stats),
			 ba_stats_total_get(report->resource_usage_stats));

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->req_resource_usage_stats), 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
      ba_stats_max_get(report->req_resource_usage_stats),
      report->resource_busy_delta_time);
    BA_ASSERT(ba_stats_count_get(report->req_resource_usage_stats)
	      == user_count);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_total_get(report->req_resource_usage_stats), 0.0);

    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
      ba_stats_total_get(report->req_resource_usage_stats),
      ba_stats_total_get(report->resource_usage_stats));

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->dis_resource_usage_stats), 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
      ba_stats_max_get(report->dis_resource_usage_stats),
      report->resource_busy_delta_time);
    BA_ASSERT(ba_stats_count_get(report->dis_resource_usage_stats)
	      == user_count);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_total_get(report->dis_resource_usage_stats), 0.0);

    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
      ba_stats_total_get(report->dis_resource_usage_stats),
      ba_stats_total_get(report->resource_usage_stats));

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->scaled_resource_usage_stats), 1.0);
    BA_ASSERT(ba_stats_count_get(report->scaled_resource_usage_stats) ==
	      user_count);

    // not checking billable_resources_usage_stats

    BA_ASSERT_REAL_EQUAL(report->resource_busy_first_abs_time,
			 ba_events_virtual_clock_initial_abs_time_get());

    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->resource_busy_last_abs_time,
				      report->virtual_clock_delta_time);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->
					 resource_busy_range_delta_time,
					 report->resource_busy_delta_time);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->resource_busy_delta_time,
					 0.0);
#if !SAVE_MEMORY
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->resource_busy_delta_time,
				      report->task_set_alive_delta_time);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->resource_busy_delta_time,
				      report->task_set_ready_tasks_delta_time);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->resource_busy_delta_time +
				      report->no_task_sets_delta_time,
				      report->virtual_clock_delta_time);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->resource_busy_delta_time +
				      report->non_work_conserving_delta_time,
				      report->virtual_clock_delta_time);
#endif

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->resource_idle_delta_time,
					 0.0);
    if(max_req_dis_tasks_per_user == 1 && user_count == 1) {
	BA_ASSERT_REAL_EQUAL(report->resource_idle_delta_time, 0.0);
    }

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->
					 non_work_conserving_delta_time,
					 0.0);
    if(!(ba_options_treat_req_dis_diff_get() &&
	 (ba_options_dis_sched_behavior_get() == BA_SCHED_BEHAVIOR_NOOP ||
	  ba_options_dis_sched_behavior_get() == BA_SCHED_BEHAVIOR_RFCFS ||
	  ba_options_dis_sched_behavior_get() == BA_SCHED_BEHAVIOR_RSRPT))) {
#if !SAVE_MEMORY
	BA_ASSERT_REAL_EQUAL(report->non_work_conserving_delta_time, 0.0);
#endif
    }

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->load, 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->load, 1.0);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->load,
					 report->req_load);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->req_load, 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->req_load, 1.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
        report->req_load, report->req_load_ignoring_idle_delta_time);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
        report->req_load_ignoring_idle_delta_time, 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
        report->req_load_ignoring_idle_delta_time, 1.0);
}

static
void assert_tasks(void)
{
    uint32_t max_req_dis_tasks_per_user;
    uint32_t user_count;

    max_req_dis_tasks_per_user =
	ba_behavior_max_req_dis_tasks_per_user_get();

    user_count = ba_behavior_num_users_get();
    BA_ASSERT(user_count == report->user_count);
    BA_ASSERT(user_count >= report->num_hit_req_dis_limit);

    BA_ASSERT(ba_stats_min_get(report->task_sets_stats) >= 1);
    BA_ASSERT(ba_stats_count_get(report->task_sets_stats) == user_count);
    
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->task_set_alive_delta_time,
					 0.0);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->task_set_alive_delta_time,
					 report->
					 task_set_ready_tasks_delta_time);
#if !SAVE_MEMORY
    BA_ASSERT_REAL_EQUAL(report->task_set_alive_delta_time +
			 report->no_task_sets_delta_time,
			 report->virtual_clock_delta_time);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->task_set_alive_delta_time,
					 report->
					 non_work_conserving_delta_time);
#endif

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->no_task_sets_delta_time,
					 0.0);
#if !SAVE_MEMORY
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->no_task_sets_delta_time +
				      report->
				      non_work_conserving_delta_time,
				      report->virtual_clock_delta_time);
#endif

    if(user_count == 1) {
	BA_ASSERT_REAL_EQUAL(report->no_task_sets_delta_time, 0.0);
    }

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->
					 task_set_ready_tasks_delta_time,
					 0.0);
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(report->
				      task_set_ready_tasks_delta_time +
				      report->no_task_sets_delta_time,
				      report->virtual_clock_delta_time);
    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(report->
					 task_set_ready_tasks_delta_time,
					 report->
					 non_work_conserving_delta_time);

    BA_ASSERT(ba_stats_min_get(report->disclosed_tasks_stats) >= 1);
    if(max_req_dis_tasks_per_user) {
	BA_ASSERT(ba_stats_max_get(report->disclosed_tasks_stats) <=
		  max_req_dis_tasks_per_user);
    }
    BA_ASSERT(ba_stats_count_get(report->disclosed_tasks_stats) ==
	      user_count);

    BA_ASSERT(ba_stats_min_get(report->disclosed_only_tasks_stats) >= 0);
    if(max_req_dis_tasks_per_user) {
	BA_ASSERT(ba_stats_max_get(report->disclosed_only_tasks_stats) <=
		  max_req_dis_tasks_per_user);
    }
    BA_ASSERT(ba_stats_count_get(report->disclosed_only_tasks_stats) ==
	      user_count);

    if(!ba_options_virtual_clock_terminate_get()) {
	BA_ASSERT(ba_stats_min_get(report->executed_tasks_stats) >= 1);
    }
    if(max_req_dis_tasks_per_user) {
	BA_ASSERT(ba_stats_max_get(report->executed_tasks_stats) <=
		  max_req_dis_tasks_per_user);
    }
    BA_ASSERT(ba_stats_count_get(report->executed_tasks_stats) ==
	      user_count);

    BA_ASSERT(ba_stats_min_get(report->requested_tasks_stats) >= 1);
    if(max_req_dis_tasks_per_user) {
	BA_ASSERT(ba_stats_max_get(report->requested_tasks_stats) <=
		  max_req_dis_tasks_per_user);
    }
    BA_ASSERT(ba_stats_count_get(report->requested_tasks_stats) ==
	      user_count);

    if(!ba_options_virtual_clock_terminate_get()) {
	BA_ASSERT(ba_stats_min_get(report->finished_tasks_stats) >= 1);
    }
    if(max_req_dis_tasks_per_user) {
	BA_ASSERT(ba_stats_max_get(report->finished_tasks_stats) <=
		  max_req_dis_tasks_per_user);
    }
    BA_ASSERT(ba_stats_count_get(report->finished_tasks_stats) ==
	      user_count);

    BA_ASSERT(ba_stats_min_get(report->canceled_tasks_stats) >= 0);
    if(max_req_dis_tasks_per_user) {
	BA_ASSERT(ba_stats_max_get(report->canceled_tasks_stats) <=
		  max_req_dis_tasks_per_user);
    }
    BA_ASSERT(ba_stats_count_get(report->canceled_tasks_stats) ==
	      user_count);

    if(!ba_options_virtual_clock_terminate_get()) {
	/* when using the virtual clock to terminate, this test is
	   wrong b/c canceled tasks will include requested tasks and
	   because not all requested tasks will finish. */
	BA_ASSERT(ba_stats_equal(report->disclosed_only_tasks_stats,
				 report->canceled_tasks_stats));
	BA_ASSERT(ba_stats_equal(report->requested_tasks_stats,
				 report->finished_tasks_stats));
    }

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->service_delta_time_stats), 0.0);
    BA_ASSERT(ba_stats_count_get(report->service_delta_time_stats) ==
	      user_count);
	      
    if(ba_stats_non_empty(report->think_delta_time_stats)) {
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
          ba_stats_min_get(report->think_delta_time_stats), 0.0);
#if !SAVE_MEMORY
	BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
			  ba_stats_max_get(report->think_delta_time_stats),
					  report->task_set_alive_delta_time);
#endif
    }
}

static
void assert_vrt(void)
{
    uint32_t max_req_dis_tasks_per_user;
    uint32_t user_count;

    max_req_dis_tasks_per_user =
	ba_behavior_max_req_dis_tasks_per_user_get();

    user_count = ba_behavior_num_users_get();
    if(!ba_options_virtual_clock_terminate_get()) {
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
	    ba_stats_min_get(report->vrt_delta_time_stats), 0.0);
#if !SAVE_MEMORY
	BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
			  ba_stats_max_get(report->vrt_delta_time_stats),
					  report->task_set_alive_delta_time);
#endif
	BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
	    ba_stats_min_get(report->vrt_delta_time_stats),
	    ba_stats_min_get(report->user_vrt_delta_time_stats));
	BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
	    ba_stats_max_get(report->vrt_delta_time_stats),
	    ba_stats_max_get(report->user_vrt_delta_time_stats));
	BA_ASSERT(ba_stats_count_get(report->vrt_delta_time_stats)
		  == ba_stats_total_get(report->finished_tasks_stats));
    }

    if(!ba_options_virtual_clock_terminate_get()) {
	BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
          ba_stats_min_get(report->vsd_stats), 0.0);
	BA_ASSERT(ba_stats_count_get(report->vsd_stats)
		  == ba_stats_total_get(report->finished_tasks_stats));
    }

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->user_vrt_delta_time_stats), 0.0);
#if !SAVE_MEMORY
    BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
		      ba_stats_max_get(report->user_vrt_delta_time_stats),
				      report->task_set_alive_delta_time);
#endif
    BA_ASSERT(ba_stats_count_get(report->user_vrt_delta_time_stats)
	      == user_count);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->user_vsd_stats), 0.0);
    BA_ASSERT(ba_stats_count_get(report->user_vsd_stats) == user_count);

    BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(
      ba_stats_min_get(report->deadlines_met_stats), 0.0);
    if(max_req_dis_tasks_per_user) {
	BA_ASSERT_REAL_LESS_THAN_OR_EQUAL(
	    ba_stats_max_get(report->deadlines_met_stats),
	    max_req_dis_tasks_per_user);
    }
    BA_ASSERT(ba_stats_count_get(report->deadlines_met_stats) == user_count);
}

//////////////////////////////////////////////////////////////////////

extern void
ba_report_output(void)
{
    process_report();

    print_proem();
    print_scheduler_params();
    print_user_behavior();
    print_scheduler();
    print_resources();
    print_tasks();
    print_vrt();

    assert_scheduler();
    assert_resources();
    assert_tasks();
    assert_vrt();
}

extern ba_bool
ba_report_warmup_met(void)
{
    // not a TYPES_REAL compare, huh.  XXX
    if(ba_events_virtual_clock_abs_time_get() >= warmup_abs_time) {
	return BA_TRUE;
    } else {
	return BA_FALSE;
    }
}

extern ba_real
ba_report_warmup_abs_time_get(void)
{
    return warmup_abs_time;
}

extern ba_status
ba_report_init(void)
{
    report = ba_mem_malloc(sizeof(report_t));

    // scheduler
    report->virtual_clock_time_period = 0;
    report->virtual_clock_delta_time = 0.0;
    report->all_candidates_rm_stats = ba_stats_running_mean_new();
    report->req_candidates_rm_stats = ba_stats_running_mean_new();
    report->dis_candidates_rm_stats = ba_stats_running_mean_new();
    report->weighed_all_candidates_rm_stats =
	ba_stats_running_mean_new();
    report->weighed_req_candidates_rm_stats =
	ba_stats_running_mean_new();
    report->weighed_dis_candidates_rm_stats =
	ba_stats_running_mean_new();

    // resources
    report->resource_usage_stats = ba_stats_new();
    report->req_resource_usage_stats = ba_stats_new();
    report->dis_resource_usage_stats = ba_stats_new();
    report->scaled_resource_usage_stats = ba_stats_new();
    report->billable_resources_usage_stats = 0; // pointer to one of the above
    report->resource_busy_periods = ba_time_periods_new();
    report->resource_busy_delta_time = 0.0;
    report->resource_idle_delta_time = 0.0;
    report->non_work_conserving_periods = ba_time_periods_new();
    report->non_work_conserving_delta_time = 0.0;
    report->load = 0.0;
    report->req_load = 0.0;
    report->req_load_ignoring_idle_delta_time = 0.0;

    // tasks and task sets
    report->user_count = 0;
    report->num_hit_req_dis_limit = 0;
    report->task_sets_stats = ba_stats_new();
    report->task_set_alive_periods = ba_time_periods_new();
    report->task_set_alive_delta_time = 0.0;
    report->no_task_sets_periods = ba_time_periods_new();
    report->no_task_sets_delta_time = 0.0;
    report->task_set_ready_tasks_periods = ba_time_periods_new();
    report->task_set_ready_tasks_delta_time = 0.0;
    report->disclosed_tasks_stats = ba_stats_new();
    report->disclosed_only_tasks_stats = ba_stats_new();
    report->executed_tasks_stats = ba_stats_new();
    report->requested_tasks_stats = ba_stats_new();
    report->finished_tasks_stats = ba_stats_new();
    report->canceled_tasks_stats = ba_stats_new();
    report->service_delta_time_stats = ba_stats_new();
    report->think_delta_time_stats = ba_stats_new();

    // visible response time
    report->vrt_delta_time_stats = ba_stats_new();
    report->vsd_stats = ba_stats_new();
    report->user_vrt_delta_time_stats = ba_stats_new();
    report->user_vsd_stats = ba_stats_new();
    report->deadlines_met_stats = ba_stats_new();

    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_report_destroy(void)
{
    // scheduler
    if(report->virtual_clock_time_period) {
	ba_time_period_delete(report->virtual_clock_time_period);
    }
    ba_stats_running_mean_delete(report->all_candidates_rm_stats);
    ba_stats_running_mean_delete(report->req_candidates_rm_stats);
    ba_stats_running_mean_delete(report->dis_candidates_rm_stats);
    ba_stats_running_mean_delete(report->weighed_all_candidates_rm_stats);
    ba_stats_running_mean_delete(report->weighed_req_candidates_rm_stats);
    ba_stats_running_mean_delete(report->weighed_dis_candidates_rm_stats);

    // resources
    ba_stats_delete(report->resource_usage_stats);
    ba_stats_delete(report->req_resource_usage_stats);
    ba_stats_delete(report->dis_resource_usage_stats);
    ba_stats_delete(report->scaled_resource_usage_stats);
    // no need to delete billable_resources_usage_stats, as it's an
    // alias to one of the above
    ba_time_periods_delete(report->resource_busy_periods);
    ba_time_periods_delete(report->non_work_conserving_periods);

    // tasks and task sets
    ba_stats_delete(report->task_sets_stats);
    ba_time_periods_delete(report->task_set_alive_periods);
    ba_time_periods_delete(report->no_task_sets_periods);
    ba_time_periods_delete(report->task_set_ready_tasks_periods);
    ba_stats_delete(report->disclosed_tasks_stats);
    ba_stats_delete(report->disclosed_only_tasks_stats);
    ba_stats_delete(report->executed_tasks_stats);
    ba_stats_delete(report->requested_tasks_stats);
    ba_stats_delete(report->finished_tasks_stats);
    ba_stats_delete(report->canceled_tasks_stats);
    ba_stats_delete(report->service_delta_time_stats);
    ba_stats_delete(report->think_delta_time_stats);

    // visible response time
    ba_stats_delete(report->vrt_delta_time_stats);
    ba_stats_delete(report->vsd_stats);
    ba_stats_delete(report->user_vrt_delta_time_stats);
    ba_stats_delete(report->user_vsd_stats);
    ba_stats_delete(report->deadlines_met_stats);

    ba_mem_free(report);

    return BA_STATUS_SUCCESS;
}

extern void
ba_report_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_report_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
