#include <ba_config.h>
#include <ba_sim.h>

static void add_idle_event_helper(ba_real idle_delta_time);
static gint find_resource_busy_events(gconstpointer event_in,
				      gconstpointer unused);
static ba_event_callback_t idle_callback;

static ba_task_t *currently_running_task;

// not needed for anything, but later might use in ba_report in an
// assert or something...
static ba_real total_req_resource_usage;

extern void
ba_resource_print(void)
{
    ba_print_info("resource:\n");
    ba_print_info("  currently_running_task: %p", currently_running_task);
    ba_print_info("  total_req_resource_usage: " BA_REAL_FMT "\n",
		  total_req_resource_usage);
}

extern ba_bool
ba_resource_is_idle(void)
{
    if(currently_running_task) {
	return BA_FALSE;
    } else {
	return BA_TRUE;
    }
}

extern void
ba_resource_busy_callback(ba_event_kind_t kind,
			  ba_real delay_delta_time,
			  void *closure)
{
    BA_ASSERT(kind == BA_EVENT_KIND_RESOURCE_BUSY);
    BA_ASSERT(closure);
    ba_task_t *task;
    ba_real remaining_delta_time;

    BA_ASSERT(currently_running_task == 0);

    task = (ba_task_t *)closure;

    remaining_delta_time = ba_task_remaining_delta_time_get(task);
    currently_running_task = task;

    ba_task_running_set(task);

#if BA_SIM_INSPECT
    ba_print_info("resource running user %d's task %d for at most "
		  BA_REAL_LOW_PREC_FMT " at time "
		  BA_REAL_LOW_PREC_FMT "\n",
		  ba_user_id_get(ba_task_user_get(task)),
		  ba_task_id_get(task),
		  ba_task_remaining_delta_time_get(task),
		  ba_events_virtual_clock_abs_time_get());
#endif

    add_idle_event_helper(remaining_delta_time);
}

extern void
ba_resource_preempt(void)
{
    add_idle_event_helper(BA_EVENTS_NOW);
}

static void
add_idle_event_helper(ba_real idle_delta_time)
{
    ba_events_remove_events(0, find_resource_busy_events);

    ba_events_replace_later_event(BA_EVENT_KIND_RESOURCE_IDLE, idle_delta_time,
				  idle_callback, 0);
}

static gint
find_resource_busy_events(gconstpointer event_in,
			  gconstpointer unused)
{
    BA_ASSERT(event_in);
    ba_event_t *event;
    
    event = (ba_event_t *)event_in;

    if(ba_event_kind_get(event) == BA_EVENT_KIND_RESOURCE_BUSY) {
	return 0;
    } else {
	return 1;
    }
}

static void
idle_callback(ba_event_kind_t kind,
	      ba_real delay_delta_time,
	      void *closure)
{
    // closure unused
    ba_task_t *task;
    ba_real remaining_delta_time;

    BA_ASSERT(kind == BA_EVENT_KIND_RESOURCE_IDLE);

    // see, there wouldn't be a currently_running_task if the system
    // just started or if the system was idle for a while and then a
    // new task_set came into existence.
    if(currently_running_task) {
	ba_real virtual_clock_abs_time;
	ba_real start_abs_time, end_abs_time;
	ba_user_t *user;

	task = currently_running_task;
	currently_running_task = 0;

#if BA_SIM_INSPECT
	ba_print_info("resource going idle after running user %d's task %d "
		      "at time " BA_REAL_LOW_PREC_FMT "\n",
		      ba_user_id_get(ba_task_user_get(task)),
		      ba_task_id_get(task),
		      ba_events_virtual_clock_abs_time_get());
#endif

	ba_task_running_unset(task);

	// divine all the times i need
	virtual_clock_abs_time = ba_events_virtual_clock_abs_time_get();
	start_abs_time = virtual_clock_abs_time - delay_delta_time;
	end_abs_time = virtual_clock_abs_time;

	// update task info
	ba_task_add_completed_delta_time(task, delay_delta_time);

	// new code begin
	if(ba_report_warmup_met()) {
	    ba_real warmup_delay_delta_time;
	    ba_real warmup_abs_time;

	    warmup_abs_time = ba_report_warmup_abs_time_get();
	    warmup_delay_delta_time = delay_delta_time;
	    if(ba_types_real_less_than_or_equal(start_abs_time,
						warmup_abs_time)) {
		// ahh, the resource run started before warm up.  so we
		// overcounted.  let's fix that.
		warmup_delay_delta_time -= (warmup_abs_time - start_abs_time);
		
		BA_ASSERT_REAL_GREATER_THAN_OR_EQUAL(warmup_delay_delta_time,
						     0.0);
	    }
	    ba_task_add_completed_warmup_delta_time(task,
						    warmup_delay_delta_time);
	}
	// new code end

	remaining_delta_time = ba_task_remaining_delta_time_get(task);

	// update user info
	user = ba_task_user_get(task);
	ba_user_add_resource_usage(user, delay_delta_time, start_abs_time,
				   end_abs_time, task);

	// update resource info
	ba_report_resource_busy_start_end(start_abs_time, end_abs_time);

	// see if task is done
	if(ba_types_real_equal(remaining_delta_time, 0.0)) {
	    ba_task_set_t *task_set;

	    task_set = ba_task_task_set_get(task);
	    ba_task_set_last_task_executed_abs_time_set(task_set,
							virtual_clock_abs_time);

	    ba_user_add_executed_task(user, task);
	}
    }

    ba_sched_choose();
}

extern void
ba_resource_add_req_resource_usage(ba_real req_resource_usage)
{
    total_req_resource_usage += req_resource_usage;
}

extern ba_status
ba_resource_init(void)
{
    currently_running_task = 0;

    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_resource_destroy(void)
{
    // should make a call to ba_events to confirm there are NO idle or
    // busy events leftover...  XXX

    return BA_STATUS_SUCCESS;
}

extern void
ba_resource_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_resource_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
