#include <ba_config.h>
#include <ba_sim.h>

static GSList *users;

static ba_id_gen_t *id_gen;

static ba_event_callback_t vc_terminate;
static uint32_t num_users;
static ba_real virtual_clock_terminate;
static ba_bool virtual_clock_terminate_done;
static uint32_t num_queued_users;
static ba_rv_dist_t *user_arrivals_dist;

static void create_user(void);
static void do_bounded_vc_simulation(void);
static void add_user_for_unbounded_vc(void);
static gint find_users_add_user_events(gconstpointer event_in,
						 gconstpointer unused);
static gint find_users_events(gconstpointer event_in,
				   gconstpointer unused);
static ba_event_callback_t add_user_for_unbounded_vc_callback;

static ba_trace_gen_rv_t gen_new_user;

static ba_bool first_user;

extern GSList *
ba_users_get(void)
{
    return users;
}

extern void
ba_users_start(void)
{
    if(virtual_clock_terminate) {
	BA_ASSERT(virtual_clock_terminate_done == BA_FALSE);
	do_bounded_vc_simulation();
    } else {
	BA_ASSERT(num_queued_users != num_users);
	add_user_for_unbounded_vc();
    }
}

static void
do_bounded_vc_simulation(void)
{
    uint32_t i;

    ba_events_add_event(BA_EVENT_KIND_USERS_VC_TERMINATE,
			virtual_clock_terminate, vc_terminate, 0);

    for(i = 0; i < num_users; i++) {
	create_user();
    }
}

static void
add_user_for_unbounded_vc(void)
{
    ba_real add_user_delta_time;

    if(num_queued_users == num_users) {
	return;
    }

    add_user_delta_time =
	ba_trace_get_and_log_rv(BA_TRACE_ENTRY_NEW_USER, -1, -1,
				gen_new_user, 0);
    ba_events_add_event(BA_EVENT_KIND_USERS_ADD_USER,
			add_user_delta_time,
			add_user_for_unbounded_vc_callback, 0);
    num_queued_users++;
}

static void
add_user_for_unbounded_vc_callback(ba_event_kind_t kind,
					ba_real delay_delta_time,
					void *unused)
{
    BA_ASSERT(kind == BA_EVENT_KIND_USERS_ADD_USER);

    create_user();
    add_user_for_unbounded_vc();
}

static void
create_user(void)
{
    ba_user_t *user;
    ba_id_t id;

    id = ba_id_gen_next(id_gen);

    user = ba_user_new(id);
    users = g_slist_append(users, user);

#if BA_SIM_INSPECT
    ba_print_info("user %d created at time " BA_REAL_LOW_PREC_FMT "\n",
		  ba_user_id_get(user), ba_events_virtual_clock_abs_time_get());
#endif

    ba_events_add_event(BA_EVENT_KIND_USER_ACTIVATE_NEW_TASK_SET,
			BA_EVENTS_NOW, ba_user_activate_new_task_set, user);
    ba_report_users_inc_user_count();
}

static ba_real
gen_new_user(void *generic_data)
{
    BA_ASSERT(generic_data == 0); // generic_data unused.
    ba_real rv;

    if(first_user) {
	first_user = BA_FALSE;
	return BA_EVENTS_NOW;
    }

    rv = ba_rv_dist(user_arrivals_dist);

    return rv;
}

extern void
ba_users_remove_user(ba_event_kind_t kind,
		     ba_real delay_delta_time,
		     void *closure)
{
    BA_ASSERT(kind == BA_EVENT_KIND_USERS_REMOVE_USER);
    BA_ASSERT(closure);
    ba_user_t *user;

    user = (ba_user_t *)closure;

    ba_user_deactivate_task_set(user);

    users = g_slist_remove(users, user);
    ba_user_delete(user);
}

static void
vc_terminate(ba_event_kind_t kind,
	     ba_real delay_delta_time,
	     void *closure)

{
    BA_ASSERT(kind == BA_EVENT_KIND_USERS_VC_TERMINATE);
    BA_ASSERT(closure == 0);
    GSList *node;

    ba_events_remove_events(0, find_users_add_user_events);

    // preempt must be before the following loop.  (in
    // ba_user_queue_for_destruction, remove_user events are
    // queued, and we need to make sure running tasks are not running
    // b4 removing users.)
    ba_resource_preempt();

    for(node = ba_users_get(); node; node = g_slist_next(node)) {
	ba_user_t *user;

    	user = (ba_user_t *)node->data;
	BA_ASSERT(user);
	
	ba_user_queue_for_destruction(user);
    }

    virtual_clock_terminate_done = BA_TRUE;
}

static gint
find_users_add_user_events(gconstpointer event_in,
				     gconstpointer unused)
{
    BA_ASSERT(event_in);
    ba_event_t *event;
    event = (ba_event_t *)event_in;

    if(ba_event_kind_get(event) == BA_EVENT_KIND_USERS_ADD_USER) {
	// actually, when using vc_terminate, there aren't any add user
	// events anymore...
	BA_PANIC();
	return 0;
    } else {
	return 1;
    }
}

extern void
ba_users_print(void)
{
    ba_print_info("Users:\n");
    ba_print_info("  id_gen: %p\n", id_gen);
    g_slist_foreach(users, ba_user_print_from_list, 0);
    ba_print_info("  user_arrivals_dist %p:\n", user_arrivals_dist);
    ba_rv_dist_print(user_arrivals_dist);
    ba_print_info("  first_user: ");
    ba_types_bool_print(first_user);
    ba_print_info("\n");
}

extern ba_status
ba_users_init(void)
{
    users = 0;

    id_gen = ba_id_gen_new();
    
    num_users = ba_behavior_num_users_get();
    virtual_clock_terminate = ba_behavior_virtual_clock_terminate_get();
    virtual_clock_terminate_done = BA_FALSE;
    num_queued_users = 0;
    user_arrivals_dist =
	ba_rv_dist_copy(ba_options_user_arrivals_dist_get());

    first_user = BA_TRUE;
    
    return BA_STATUS_SUCCESS;
}

extern ba_status
ba_users_destroy(void)
{
    BA_ASSERT(users == 0);

    // there's an assert in the following function as we shouldn't
    // find such events at this stage.
    ba_events_remove_events(0, find_users_events);

    g_slist_free(users); // kind of moot, since it's 0 here.

    ba_id_gen_delete(id_gen);

    ba_rv_dist_delete(user_arrivals_dist);

    return BA_STATUS_SUCCESS;
}

static gint
find_users_events(gconstpointer event_in,
		       gconstpointer unused)
{
    BA_ASSERT(event_in);
    ba_event_t *event;

    event = (ba_event_t *)event_in;

    if(ba_event_kind_get(event) == BA_EVENT_KIND_USERS_ADD_USER ||
       ba_event_kind_get(event) == BA_EVENT_KIND_USERS_REMOVE_USER) {
	BA_PANIC(); // should never happen
	return 0;
    } else {
	return 1;
    }
}

extern void
ba_users_destroy_from_hook(void *unused)
{
    ba_status status;

    status = ba_users_destroy();
    BA_ASSERT(status == BA_STATUS_SUCCESS);
}
